; ====================================================================================
; UT4 Monitor ROM assembler source code for RCA CDP18S020 development board
;
; This is a regenerated assembler source code image for the UT4 (Serial)
; monitor ROM for the RCA CDP18S020 development board. This was originally supplied
; with the CDP18S020 kit (marked CDPR512), and fits in socket U2 on the board. 
; This code will assemble to produce an exact image of the original monitor image, 
; except for a few undefined bytes (set to zero here).
;
; This image has been generated by hand editing the hex file for the UT4 monitor from
; the listing in the MPM-224 RCA COSMAC CDP18S020 manual, then using the DASMX
; disassembler to generate a (poor quality) assembler file. After much touching up
; and redefinition of labels, etc., the excellent a18.exe assembler has been used to
; regenerate a new hex file that matches the original one. The assembler source has 
; also been compared to the original MPM-224 listing to verify its integrity.
;
; The source code uses some weird and wonderful tricks within the 1802 CPU, including
; the liberal use of SEP instructions, and the more subtle next byte OUT mechanism.
; This can make understanding the program flow pretty tricky.
;
; This listing has been refined to be fully relocatable, so should be adaptable for other
; projects where a 1802-based system is connected to a serial port.
; Note that there are some expedient register setups, which may break if the code is
; arbitrarily relocated; these are marked with ###.
;
; NOTE: This routine will use the 32 byte CDP1824 RAM for register storage, if it 
; detects the presence of this RAM. However, the RAM is NOT mandatory for this monitor, 
; so it can be used in systems without this RAM fitted.
;
; Hardware serial interface is very simple, so can be replicated on boards other than
; the original CDP18S020:
; - Serial transmit output is connected to CPU Q line
; - Serial receive input is connected to CPU EF4 line
; Note that for specialised paper tape application, additional circuitry is required
; for reader start/stop control on IO address 7. Not required for normal serial usage.
; For some reason, this routine also outputs on IO address 1, but this has no effect
; on the CDP18S020 board. 
;
; NOTE: There appears to be some inconsistency between the terminal echo and direct
; sending of characters. In the echo mode, only the bottom 7 bits are output, followed
; by the stop bit (terminal must run in 7n1 or 7n2 mode). But the direct output 
; routine appears to send the full 8 data bits, meaning the top bit is low.
; This would suggest 8n1 mode, as the stop bit would not be valid for 7n1 or 7n2.
; This obviously worked with terminal hardware in its day, but it is not clear how!
; Also note that on receive, only the 7 bottom data bits are processed, and the top
; (8th) data bit is always set to zero internally.
;
; ====================================================================================

; Ensure we are building this for a 1802 CPU without any enhancements
        CPU     1802

; ====================================================================================
; Useful definitions for this module

; ------------------------------------------------------------------------------------
; Base register definitions for this module
; Where registers have a specific use in the monitor, this is indicated in comments
; NOTE: Registers have different usages in the separate user-accessible LEDD routine

r0	        EQU	0           ; PTER - auxiliary for main routine
r1		EQU	1           ; CL, SWITCH - clobbered, mode switch
r2		EQU	2           ; ST - stack pointer reference to RAM
r3		EQU	3           ; SUB - subroutine program counter
r4		EQU	4           ; general purpose usage
r5		EQU	5           ; PC - main program counter
r6		EQU	6           ; 
r7		EQU	7           ;
r8		EQU	8           ; 
r9		EQU	9           ; 
r10		EQU	10          ;
r11		EQU	11          ;
r12		EQU	12          ; DELAY1 - program counter for DELAY1 subroutine
r13		EQU	13          ; ASL, CENTER - hex assembly for input/output
r14		EQU	14          ; AUX - time constant/record + general usage
r15		EQU	15          ; CHAR - input/output char + general usage

; ------------------------------------------------------------------------------------
; Fundamental definitions for this module

; register storage is set near the top of the special 32 byte CDP1824 register storage RAM
; note that the full 32 bytes all the way fro $8C00 to $8C1F are actually used, if present
REG_STORE	EQU	$8C1E


; ====================================================================================
; Define starting point for this ROM image in the target system
	org	$8000

; ====================================================================================
; Main entry point for the UT4 monitor program - executed whenever board is reset
; Initially, this stores all the registers in special RAM if that RAM has been found
; It uses a very arcane method of extracting the register contents, literally writing
; an instruction sequence to RAM then executing it, and storing the result in RAM
; This is more efficient that simply loading and storing register contents, but 
; far less easy to understand. 

ENTER:
; set r0 high to the main UT4 routine high address, ready for execution
	nop
	ldi	HIGH(UT4)
	phi	r0
; set r1 to point to the special RAM area for register storage, if available
	ldi	HIGH(REG_STORE)
	phi	r1
	ldi	LOW(REG_STORE)
	plo	r1
; store a special modified instruction in register r4
; this will be used as a basis for obtaining the register values
	ldi	$A0
	phi	r4
; now attempt to test for presence of special RAM area for register storage
	sex	r1                  ; use r1 for indexing
LOOP2:
; store "sep r0" instruction in memory indexed by r1, and check if wrote
	ldi	$D0                 ; set special test value
	str	r1                  ; write to memory addressed by r1
	xor                         ; xor with stored memory indexed by r1
	bnz	UT4                 ; if not written, exit to monitor
; successfully wrote, so decrement index value and check modified instruction
; this instruction will be a "plo rx" or "pli rx", to extract register value
	dec	r1                  ; decrement index value
	ghi	r4                  ; obtain modified instruction
	adi	$70                 ; check if in the 90s
	bdf	SKIP1               ; jump if already in 90s
	adi	$21                 ; if not, 8N becomes 9N
SKIP1:
	adi	$7F                 ; if so, 9N becomes 8(N-1)
	phi	r4                  ; store back in register r4
; now store modified instruction in memory indexed by r1, and check if wrote
	str	r1                  ; set modified instruction in RAM
	xor                         ; xor with stored memory indexed by r1
	bnz	UT4                 ; if not written, exit to monitor
; switch execution to modified instruction, which will then be followed by sep 0
	sep	r1
; execution returns here, with returned register value ready to be stored
; store return value in memory indexed by r1, now beyond end of original stored values
	str	r1
; decrement memory address twice, ready to store next register value, then loop round
	dec	r1
	dec	r1
	br	LOOP2

; ....................................................................................
; standard entry point for the serial UT4 monitor program
UT4:
; setup high program counter in r5 and subroutine high pointer in r3
	ghi	r0
	phi	r5
	phi	r3
; setup low program counter in r5 and transfer execution to r5
	ldi	LOW(UT4A)
	plo	r5
	sep	r5
UT4A:
; continue execution of monitor, setting up r5 indexing, disabling interrupts
	sex	r5                  ; setup r5 to index data for return
	dis                         ; disable interrupts and return
	db	$55                 ; defines r5 for both X and P
; output bit set to address decode 1 to apparently select "RCA group"
; this I/O address does not appear to be decoded on the CDP18S020 board!
	out	1
        db      $01
; wait for the initial prompt from the monitor to calibrate the baud rate
; setup timer calibration routine in r3 subroutine pointer and make the call
	ldi	LOW(TIMALC)
	plo	r3
	sep	r3

; initialisation is now complete, so start execution
START:
; setup delayed output from M(R5) in subroutine pointer 
	ldi	LOW(TYPE5D)
	plo	r3
; call output routine to transmit carriage return character
	sep	r3                  ; execute the TYPE5 subroutine
	db	$0D
ST2:
; call output routine to output linefeed followed by a * prompt character
	sep	r3                  ; execute the TYPE5 subroutine
	db	$0A
	sep	r3                  ; execute the TYPE5 subroutine
	db	$2A
IGNORE:
; prepare to input hex digits, so clear ASL hex digit store
	ldi	$00
	plo	r13
	phi	r13
; wait for a command to be input from the operator
	ldi	LOW(READAH)
	plo	r3
	sep	r3                  ; execute the READAH subroutine 
; check if the input was a "$" character (start of execution command)
	xri	$24                 ; check if $ char
	bz	DOLLAR              ; jump of start of exec command
; check if the character is a "!" after previous XOR, and also save mode
	xri	$05                 ; check if ! char
	plo	r1                  ; save converted value in mode switch
	lsz	                    ; long skip to address MATCH1 if match
; check if character is a "?" after previous XOR
	xri	$1E                 ; check for ? char
MATCH1:
; jump back to ignore if characters do not match	
        bnz	IGNORE
; at this point, we have either a "!" (mem write) or "?" (mem read) command
; read in the command arguments, which start with the "M" command char
; register r1 low stores the mode (zero for write, non-zero for read)
RDARGS:
; obtain next character and check that it is the "M" command
	sep	r3                  ; execute the READAH subroutine 
	xri	$4D                 ; check for "M" command
	bnz	SYNERR              ; jump to error handler if not valid
RD1:
; loop round until a hex character has been received
	sep	r3                  ; execute the READAH subroutine
	bnf	RD1                 ; loop round if NOT hex char
ENDHX:
; now loop round loading hex chars until non-hex char received
	sep	r3                  ; execute the READAH subroutine
	bdf	ENDHX               ; loop round while hex char
; check that the non-hex character was a space, separating address from data
	xri	$20                 ; check for space char
	bnz	SYNERR              ; jump to error handler if not valid
; now transfer loaded hex digits to r0, to act as a memory pointer
	ghi	r13
	phi	r0
	glo	r13
	plo	r0
; now check whether read or write mode has been selected
	glo	r1                  ; obtain the mode
	bz	EX1                 ; jump if write memory mode

; ....................................................................................
; Read memory command, so obtain number of bytes, ready for dumping memory
; first clear input hex digits, ready for input of number of bytes
	ldi	$00
	plo	r13
	phi	r13
RD2:
; now read in the number of bytes to be read from the operator
	sep	r3                  ; execute the READAH subroutine
	bdf	RD2                 ; loop round while hex char
; finished command, so check that a CR was entered to terminate it
	xri	$0D                 ; check for CR command termination
	bnz	SYNERR              ; jump to error handler if not valid
; read command complete, so proceed to read memory and output to console
; setup delayed output from M(R5) in subroutine pointer 
	ldi	LOW(TYPE5D)
	plo	r3
; transfer the number of bytes to read into register r1
	glo	r13
	plo	r1
	ghi	r13
	phi	r1
LINE:
; start of line, so output a linefeed initially
	sep	r3                  ; execute the TYPE5D subroutine
	db	$0A
LINE1:
; output the high byte of the address as two hex chars
	ghi	r0
	phi	r15
; setup output r15 as two hex digits in subroutine pointer 
	ldi	LOW(TYPE2)
	plo	r3
	sep	r3                  ; output 2 hex digits for r15 high
; output the low byte of the address as two hex chars
	glo	r0
	phi	r15
; setup output r15 as two hex digits in subroutine pointer 
	ldi	LOW(TYPE2)
	plo	r3
	sep	r3                  ; output 2 hex digits for r15 high
OSPC1:
; output a space separator between hex digit pairs
	sep	r3                  ; execute the TYPE5 subroutine
	db	$20
; now output a line of data bytes, up to the count or end of line (whichever is lower)
TLOOP:
; load data from memory addressed by r0 register, ready for output
	lda	r0
	phi	r15
; setup output r15 as two hex digits in subroutine pointer 
	ldi	LOW(TYPE2)
	plo	r3
	sep	r3                  ; output 2 hex digits for r15 high
; decrement byte counter in r1 low and check if non zero
	dec	r1
	glo	r1
	bnz	TL3                 ; skip next section if non zero
; decrement byte counter in r1 high and check if non zero
	ghi	r1
	bz	START               ; terminate output once byte counter hit zero
TL3:
; now check if we have reached the end of the line (16 bytes output)
	glo	r0
	ani	$0F                 ; mask out top nibble
	bnz	TL2                 ; continue if not end of line
; end of line, so output a semicolon, followed by a carriage return
	sep	r3                  ; execute the TYPE5 subroutine
	db	$3B
	sep	r3                  ; execute the TYPE5 subroutine
	db	$0D
; loop round to generate next line of the listing
	br	LINE
TL2:
; check if we should insert a space between digit pairs
	shr                         ; divide by two
	bdf	TLOOP               ; odd value, so loop back
	br	OSPC1               ; write space and loop back

; ....................................................................................
; Write memory command; reads in hex chars, writing to memory until terminating CR
EX3:
; loop round until a hex character has been received
	sep	r3                  ; execute the READAH subroutine
	bnf	EX3                 ; loop round until hex char received
EX2:
; now load second hex character, making a hex digit pair
	sep	r3                  ; execute the READAH subroutine
	bnf	SYNERR              ; abort if non-hex character
; obtain hex digit data and store at current address, incrementing after
	glo	r13
	str	r0
	inc	r0
EX1:
; load next character, and check character type to determine action
	sep	r3                  ; execute the READAH subroutine
	bdf	EX2                 ; loop back round if hex character
; check if terminating carriage return has been detected
	xri	$0D                 ; check for carriage return
	bz	START               ; reset interpreter back to start if so
EX4:
; check if comma character (means data continues on a new line)
	xri	$21                 ; check for comma char (after prev XOR)
	bz	EX3                 ; jump back to write start if so
; no check if a semicolon character (means data loads from a new address)
	xri	$17                 ; check for semicolon char (after prev XOR)
	bnz	EX1                 ; loop back to data load if NOT
EX5:
; wait for terminating CR character before starting new address load
	sep	r3                  ; execute the READAH subroutine
	xri	$0D                 ; check if CR character
	bnz	EX5                 ; loop round if not
; new line, so reset to load new address for memory write
	br	RD1

; ....................................................................................
; Outputs error indication (CR, LF, ?), and then resets interpreter to start
SYNERR:
; setup delayed output from M(R5) in subroutine pointer
	ldi	LOW(TYPE5D)
	plo	r3
; output a carriage return character
	sep	r3                  ; execute the TYPE5D subroutine
	db	$0D
; jump to output linefeed and question, then loop back to start
	lbr	FSYNER
; extra bytes, not defined in original source code
	db	$00, $00, $00, $00

; ....................................................................................
; Called when the $ command is entered, forming the start of the execution command
; The command starts with $P, followed by execution address and then CR 
DOLLAR:
; wait for next character and check if it is a "P" ($P command)
	sep	r3                  ; execute the READAH subroutine 
	xri	$50                 ; check for "P" command
	bnz	SYNERR              ; jump to error handler if not valid
D1LP:
; loop round loading hex characters until non hex detected
	sep	r3                  ; execute the READAH subroutine 
	bdf	D1LP                ; loop round while hex char
; check that terminating character was a CR, and abort if not
	xri	$0D                 ; check for terminating CR
	bnz	SYNERR              ; abort with error if not
; now transfer loaded address into r0 execution address
	ghi	r13
	phi	r0
	glo	r13
	plo	r0
; setup delayed output from M(R5) in subroutine pointer 
	ldi	LOW(TYPE5D)
	plo	r3
; write out a line feed character before starting execution
	sep	r3                  
	db	$0A
; return into user program, using register r0 for both X and P
	sex	r5                  ; setup r5 to index data for return
	ret                         ; execute return to user program
	db	$00                 ; loads r0 into X and P

; ....................................................................................
; Delay exit routine; executes return to subroutine caller, address in r3
; This means, delay can only be invoked from a subroutine, not main program
; On exit, r3 will point back at DELAY1 routine, ready for subsequent invocations
DEXIT:
	sep	r3
; main delay routine; load inner delay value off call stack (next location after sep call)
; outer delay value is governed by r14 high, which runs a countdown in r14 low
; It appears that r14 is set up as a time constant, computed when CR/LF has been received
; Delay is computed as 2(1+r14high(3*sub parm))
; Time constant is (terminal bit time /(  2-*Cosmac instruction time)) - 1
;
DELAY1:
; obtain time constant, divide by 2 and set for counter init
	ghi	r14
	shr                         ; get delay without the echo flag
	plo	r14
DELAY2:
; decrement conter, load in specified parameter byte from memory
	dec	r14
	lda	r3                  ; load in parameter for inner count (and advance)
DELAY3:
; decrement innder count and loop round until done
	smi	$01
	bnz	DELAY3
; check if outer counter has reached zero, and exit if so
	glo	r14
	bz	DEXIT
; not finished, so adjust memory pointer and loop back around
	dec	r3                  ; compensate for previous advance
	br	DELAY2
;
; ....................................................................................
; This is the main byte time calibration routine, and also sets echo/no echo mode
; Waits for LF (no echo) or CR (echo) to be received, and also sets up delay pointer
; r14 high will hold time constant in top 7 bits; lowest bit contains echo/no echo flag
; Flag will be SET for no echo, and CLEAR for echo enabled.
; In binary, CR is 00001101 (0x0D) and LF is 00001010 (0x0A)
; This routine will calibrate the time of the top four zero bits
; After bit 3 is received, it then checks bit 2 to identify CR or LF character
;
TIMALC:
; store current PC high in delay address high (for expediency)
; this could cause issues if the code layout is modified ###
	ghi	r3
	phi	r12
; zero the calibration factor in r14 and received char in r15
	ldi	$00
	plo	r14
	plo	r15
; store r12 low pointing to delay subroutine for later use
	ldi	LOW(DELAY1)
	plo	r12
WAIT1:
; wait for start bit to be received, and loop round until this occurs
	b4	WAIT1
WAIT2:
; now wait for first data bit to be received, and loop round until this occurs
	bn4	WAIT2
; wait for a small delay loop before continuing
	ldi	$03
TC2:
; subtract loop count and loop round until done
	smi	$01
	bnz	TC2
; check if data has changed previously
	glo	r15
	bnz	ZRONE               ; branch if data has changed
; check if data has changed to 0 and branch if not
	b4	INCR
; data has changed, son increment character data
	inc	r15
ZRONE:
; check if data has changed to 1 and branch if so
; this will be when the first set bit (bit 3) of the data is received
	b4	DAUX
INCR:
; increment calibration value at this point
	inc	r14
; set up for a longer delay loop and loop round again
        ldi	$07
	br	TC2
DAUX:
; bit 3 of the data has been received at this point - calibration time complete
; reduce count to balance fixed overload in calling delay
	dec	r14
	dec	r14
; obtain accumulated counter, set bottom bit and set as calibration value
	glo	r14
	ori	$01                 ; disable echo by setting bit 0 high
	phi	r14
; wait for 1.5 bit timing delay before checking flag
; we should then be in the middle of bit 2, to determine CR/LF status
	sep	r12                 ; call the DELAY1 routine
	db	$0C
; branch if line feed (no echo); continue if CR (echo)
	bn4	WAIT
; turn echo ON, by zeroing bottom bit of r14 high
	ghi	r14
	ani	$FE                 ; enable echo, by setting bit 0 low
	phi	r14
WAIT:
; wait for delay for character bit sequence to complete
	sep	r12                 ; call the DELAY1 routine
	db	$26
; return back to the caller
	sep	r5

; ....................................................................................
; Ancillary routine for the internal read processor handling and exiting
; Belongs further down, but may be expedient reason for including up here
CKDEC:
; check character for ASCII decimal input (on entry, char has $41 subtracted)
	adi	$07                 ; align to start for character 0
	bdf	NFND                ; jump if not in correct range
	adi	$0A                 ; check if numerical value below 10
	bdf	FND                 ; jump back to found if so
NFND:
; character does NOT match ASCII range, so clear DF flag and exit
; DF is returned clear if not an ASCII char; else it is returned set
	adi	$00                 ; clear the DF flag
REXIT:
; exit from the read routine, returning char in D register as well as R15
; This leaves R3 pointing back at the READAH routine for next invocation
	ghi	r15                 ; obtain r15 register in D
	sep	r5
;
; ....................................................................................
; Routine to input ASCII character from serial port, stored in 15 high
; Also processes hex characters in a special manner, shifting into r13
; Mode bit 0 clear disables paper tape, higher bits clear enable special hex mode
; On eventual return, DF set if hex char, clear if other char. 
; On exit, R3 will point to READAH ready for next invocation.
READAH:
	ldi	$00                 ; zero flag for terminal usage/special hex mode
	skp                         ; effectively skip to READ1
; ....................................................................................
; Routine to input ASCII character from serial port, stored in 15 high
; This loads in the low address of TIMALC for setting mode (very hacky, but saves a byte)
; With the address at $FE, this disables paper tape mode and the special hex processor!
; On exit, R3 will point to READAH ready for next invocation.
; If we move things around in EPROM, this would break operation big-time! ### 
READ:
        glo     r3                  ; loads in TIMALC low addr for mode setting!
	lskp			    ; skip to address READ1
; ....................................................................................
; Routine to input ASCII character from paper tape, stored in 15 high
; Also processes hex characters in a special manner, shifting into r13
; This includes control of reader via special output on PTR line
; Mode bit 0 set enabled paper tape, higher bits clear enable special hex mode
; On exit, R3 will point to READAH ready for next invocation.
; NOTE: Only 7 bit values are stored; the last (top) bit is IGNORED.
; The echo ONLY outputs 7 data bits followed by stop bit, so receiver
; must be operating in 7 bit mode (7n1 or 7n2)
;
TTYRED:
	ldi	$01                 ; set flag for paper tape usage
READ1:
; store terminal/paper tape flag in r15 low
	plo	r15
READ2:
; initialise the input byte, to allow detection of last bit entered
; this actually only processes 7 bits of data in the sequence
	ldi	$80                 ; top bit will eventually shift out
	phi	r15
	sex	r3
; obtain paper tape/serial selection, and determine the mode
	glo	r15
	shr
	bnf	RCONT1               ; jump over if NOT paper tape
; turn on the teletype reader via special PTR signal (P1 pin 20)
; this relies on special decode/latch circuitry on the CDP18S020 board
	out	7
        db      $80
RCONT1:
; wait for the end of the last stop bit
	bn4	RCONT1              ; loop round while bit set
TTY1:
; wait for start bit to be detected (input from EF4)
	b4	TTY1                ; loop round if no start bit
; start bit detected, so wait for half bit delay and check again
	sep	r12                 ; call the DELAY1 subroutine
	db	$02
	b4	TTY1                ; loop round if no start bit
; obtain paper tape/serial selection, and determine the mode
	glo	r15
	shr
	bnf	NOBIT               ; jump over if NOT paper tape
; turn off the teletype reader via special PTR signal (P1 pin 20)
; this relies on special decode/latch circuitry on the CDP18S020 board
	out	7
        db      $40
NOBIT:
; reset X selection and wait for short delay
	sex	r2
	nop
BIT:
; check whether or not we are in echo mode (bit 0 set of r14 high)
	ghi	r14
	shr
	bdf	NOECHO              ; jump to no echo if off
; echo the incoming bit directly to the output bit
	b4	OUTBIT              ; jump if bit is set
	seq                         ; set Q data output high
	br	NOECHO              ; skip over next bit
OUTBIT:
	req                         ; reset Q data output low
	nop                         ; balance timing with delay
NOECHO:
; wait for a single bit period delay
	sep	r12                 ; call the DELAY1 routine
	db	$07
	nop                         ; extra delay steps
	nop
; shift the character right one bit, and check if finished
; once top bit processed, jumps out without storing it (so will be zero)
	ghi	r15
	shr
	phi	r15
	bdf	NEXT                ; jump if top bit shifted out
; preset bit set and check if input bit was not set
	ori	$80                 ; temp mask in bit set
	bn4	NOBIT               ; loop back input bit not set
; input bit set, so store updated character with masked in bit
	phi	r15                 ; store character
	br	BIT                 ; loop back for next bit

; input character fully loaded (top bit zero), so we now process it
NEXT:
; echoes the stop bit of the data stream after only 7 data bits
; this is done even when echo is disabled, to ensure correct state
	req                         ; reset Q to output stop bit
; check if character was a null, and loop round if so
	bz	READ2               
; check entry mode, and exit if normal read mode
	glo	r15                 ; check read entry mode
	bnz	REXIT               ; exit if non-zero

; special hex character processor enabled, so check if hex character
	ghi	r15                 ; obtain the character
	smi	$41                 ; subtract letter A
	bnf	CKDEC               ; jump if numerical value
; letter hex value, so check if in range A to F
	smi	$06                 ; effective subtract letter F
	bdf	NFND                ; jump if outside range
FND:
; character is inside hex range, so first shift value up one nibble
	shl
	shl
	shl
	shl
; add in bit to ensure loop operates for correct cycles to shift nibble
	adi	$08                 ; set top bit of bottom nibble
	shl                         ; shift bit into carry, ready for update
FND1:
; temp store raw value in r14, then obtain r13 value and shift up
	plo	r14                 ; temp store value
	glo	r13                 
	shlc                        ; shift in bit from char
	plo	r13
	ghi	r13
	shlc                        ; shift in bit from low reg
	phi	r13
; obtain shifted value, and check if more work is required
; due to special bit set, will loop round until all 4 bits shifted
	glo	r14                 ; retrieve stored value
	shl                         ; detect if any data remains
	bnz	FND1                ; loop round if so
; finished, so exit from the routine, DF will be 1 due to shift
; on return, DF set means hex digit was loaded in
	br	REXIT
; extra byte, not defined in original source code
	db	$00

; ....................................................................................
; Waits for 1.5 bit delay before outputting ASCII char at M(R5), then inc R5
; R5 is the program counter, so will load byte after the original SEP call to this routine
; On exit, r3 points back to main TYPE5 routine, ready for next invocation
;
TYPE5D:
; execute the 1.5 bit timing delay before proceeding to output routine
	sep	r12                 ; call the DELAY1 subroutine
	db	$17
	skp			    ; skip over to address TYPE5
TEXIT:
; exit from the character output routine, continuing execution
; R3 will point to the main TYPE5 routine, ready for subsequent invocations
	sep	r5

; ....................................................................................
; Outputs ASCII char at memory addressed by R5, then increments R5
; R5 is the program counter, so will load byte after the original SEP call to this routine
; NOTE: This routine does appear to echo the full 8 data bits of each character, but
; this is NOT consistent with the echo function which only outputs 7 data bits.
;
TYPE5:
; obtain character from memory addressed by r5, and jump to common output
	lda	r5                  ; load from r5 address, advance address
	skp			    ; effectively skip to address TY1

; ....................................................................................
; Outputs ASCII char at memory addressed by R6, then increments R6
; On exit, r3 points back to main TYPE5 routine, ready for next invocation
;
TYPE6:
; obtain character from memory addressed by r6, and jump to common output
	lda	r6                  ; load from r6 address, advance address
	skp			    ; effectively skip to address TY1

; ....................................................................................
; Outputs ASCII char contained in register r15 high
; On exit, r3 points back to main TYPE5 routine, ready for next invocation
;
TYPE:
; obtain data directly from r15 high, and continue to common code
	ghi     r15
TY1:
; store in temp character and check if line feed
	plo	r14                 ; store in temp char 
	xri	$0A                 ; check for line feed
	bnz	TY2                 ; jump if NOT line feed
; line feed character, so set number of bits plus number of nulls
	ldi	$8B                 ; top nibble specifies null count
	br	TY3                 ; jump to common output routine

; ....................................................................................
; Outputs two hex digits contained in register r15 high
; On exit, r3 points back to main TYPE5 routine, ready for next invocation
;
TYPE2:
; obtain hex data from r15 high
	ghi	r15                 ; obtain two hex digits in r15 high
TY4:
; shift top digit to bottom digit and convert to hex
; note that bottom digit is handled later by HEX2 entry point
	shr                         ; shift first hex digit to right
	shr
	shr
	shr
; convert the single digit value to hex
	adi	$F6                 ; attempt to convert to hex value
	bnf	LHEX1               ; branch if hex letter code
	adi	$07                 ; otherwise add net 37
LHEX1:
	smi	$C6                 ; adjust down to numerical value
	plo	r14                 ; store in temp char ready for output
; load number of bits to output for two hex chars
	ldi	$1B                 ; top nibble indicates two chars
	lskp			    ; long skip to address TY3
TY2:
; load number of bits to output for single char
	ldi	$0B
TY3:
; save main tally value ready for data output
	plo	r15
BEGIN:
; output the start bit, and transfer character to r13 for output
	seq                         ; set Q data output high
	glo	r14                 ; obtain temp character value
	plo	r13                 ; store ready for output
PREBIT:
; wait for one bit time for the current bit
	sep	r12                 ; call the DELAY1 subroutine
	db	$07
; decrement bit counter, then shift bottom bit into carry flag
; we ensure that high bits are shifted in, so the form the stop bit at the end
	dec	r15                 ; decrement bit counter
	sd                          ; preset DF flag for shift
	glo	r13                 ; get shifted char value
	shrc                        ; shift data out into DF, top bit set from DF
	plo	r13                 ; store shifted character value
; use DF flag to determine whether to set or clear bit output
	bdf	OUT1B               ; branch if data bit set
	seq                         ; set Q data output high (low data bit)
	br	OUTCNT1             ; skip over next section
OUT1B:
	req                         ; reset Q data output low (high data bit)
	nop                         ; balance timing
OUTCNT1:
; obtain bit counter and check if reached end
	glo	r15                 ; obtain the bit counter
	ani	$0F                 ; mask off upper bits
	nop                         ; balance timing
	nop                         ; balance timing
	bnz	PREBIT              ; loop round until all bits sent
NXCHAR:
; Setup for next character, but exit if none required
	glo	r15                 ; obtain loop counter
	adi	$FB                 ; add on offset
	plo	r15                 ; store loop counter
	bnf	TEXIT               ; exit of no more chars
; check for alternative counter values
	smi	$1B                 ; check for special values
	bz	TEXIT               ; exit if just list null
	bnf	HEX2                ; jump if just first hex digit
; prepere to type the null character
	ldi	$00                 ; specify null character
	br	HX22                ; jump to common code
HEX2:
; two hex chars, so now extract the low character value
	ghi	r15                 ; obtain two hex digits in r15 high
	ani	$0F                 ; mask out upper digit
	adi	$F6                 ; attempt to convert to hex value
	bnf	LHEX2               ; branch if hex letter code
	adi	$07                 ; otherwise add net 37
LHEX2:
	smi	$C6                 ; adjust down to numerical value
HX22:
; store in temp character and jump to common routine
	plo	r14                 ; store in temp char ready for output
	br	BEGIN

; ....................................................................................
; Outputs a linefeed and question mark - should be called if error occurred
; On entry, r3 should be setup to point to TYPE5D
; This really forms part of the monitor program higher up in the ROM
;
FSYNER:
; output linefeed, followed by question mark
	sep	r3                  ; execute the TYPE5D subroutine
	db	$0A
	sep	r3                  ; execute the TYPE5D subroutine
	db	$3F
; jump back to monitor start entry point
	lbr	START
	db	$00

; ====================================================================================
; end of monitor program listing assembly
    END

; ====================================================================================
; ====================================================================================
